/* model.c++ presents data model 
             (Inventor model and RAPID model)
             Physical Based Simulation, ETH Zuerich
             Rene Mueller, April 14th 2003 */

#include "model.h"
#include "unistd.h"
#include <iostream>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>

using namespace std;

//! read new model from file 
Model::Model(const char* filename)
{
   rootModel = new SoSeparator;
   material = new SoMaterial;
   rootModel->addChild(material);
   rapidModel = new RAPID_model;

   // load model from IV file 
   SoInput *input = new SoInput;
   if (!input->openFile(filename)) {
      cerr << "File '" << filename << "' could not be opened." << endl;
      exit(1);
   }
   SoDB::init();
   
   model = SoDB::readAll(input); // read entire object
   if (model == NULL) {
      cerr << "Error on read of '" << filename << "'" << endl;
      exit(1);
   }
   rootModel->addChild(model);

   // fill up RAPID structure
   
   SoCoordinate3 *coords = NULL;
   SoIndexedFaceSet *faces = NULL;
   SoIndexedTriangleStripSet* triangles = NULL;
   // find first SoCoordinate3 node 
   for (int i=0; i<model->getNumChildren(); i++) {
      SoNode* child = model->getChild(i);
      if (child->isOfType(SoCoordinate3::getClassTypeId())) {
         coords = dynamic_cast<SoCoordinate3*>(child); 
         cout << "reading model '" << filename << "': node " << child->getTypeId().getName().getString() << " read." << endl;
      } else if (child->isOfType(SoIndexedFaceSet::getClassTypeId())) {
         faces = dynamic_cast<SoIndexedFaceSet*>(child);
         cout << "reading model '" << filename << "': node " << child->getTypeId().getName().getString() << " read." << endl;
      } else if (child->isOfType(SoIndexedTriangleStripSet::getClassTypeId())) {
         triangles = dynamic_cast<SoIndexedTriangleStripSet*>(child);
         cout << "reading model '" << filename << "': node " << child->getTypeId().getName().getString() << " read." << endl;
      } else {
         cout << "reading model '" << filename << "': node " << child->getTypeId().getName().getString() << " ignored" << endl;
      }
   }
   if (triangles == NULL) {
      // we are using SoCoordinate3 and SoIndexedFaceSet
      assert(coords);
      assert(faces);
   
      // verify that face simlices are only triangles 
      // and create RAPID data model
      int vertices = 0, faceNumber = 0;
      double a[3], b[3], c[3];      // vertex data
      for (int i=0; i<faces->coordIndex.getNum(); i++) {
         int element = faces->coordIndex[i];
         if (element == -1) {
            if (vertices != 3) {
               cout << "Model '" << filename << +"' does not only contain triangular simplices! (count " << vertices << ")" << endl;
               exit(1);
            } else {
               rapidModel->AddTri(a, b, c, faceNumber);
               faceNumber++;
               vertices = 0;
            }
         } else {
            if (vertices == 0) {
               // first point
               SbVec3f point = coords->point[element];
               a[0] = point[0]; a[1] = point[1]; a[2] = point[2];
            } else if (vertices == 1) {
               // second point
               SbVec3f point = coords->point[element];
               b[0] = point[0]; b[1] = point[1]; b[2] = point[2];
            } else {
               // third point
               SbVec3f point = coords->point[element];
               c[0] = point[0]; c[1] = point[1]; c[2] = point[2];
            }
            vertices++;
         }
      }
   } else {
      // we are using SoIndexedTriangleStripSet
      assert(triangles);
      SoVertexProperty* prop = dynamic_cast<SoVertexProperty*>(triangles->vertexProperty.getValue());
      int numVertices = prop->vertex.getNum();
      // create RAPID model
      double p[2][3], a[3]; 
      bool stripReady = false; 
      int lastPtr = -1, faceNumber = 0;
      for (int i=0; i<triangles->coordIndex.getNum(); i++) {
         int element = triangles->coordIndex[i];
         if ((element != -1) && ((element <0) || (element > numVertices))) {
            cerr << "invalid vertex index " << element << "in triangle strip set." << endl;
            exit(1);
         } 
         if (element != -1) {
            // end of strip not reached yet
            if (stripReady) {
               a[0] = prop->vertex[element][0]; a[1] = prop->vertex[element][1]; 
               a[2] = prop->vertex[element][2];
               rapidModel->AddTri(p[lastPtr], p[(lastPtr+1)%2], a, faceNumber);
               lastPtr = (lastPtr%2)+1;
               faceNumber++;
               p[lastPtr][0] = a[0]; p[lastPtr][1] = a[1]; p[lastPtr][2] = a[2];
            } else {
               lastPtr = (lastPtr+1)%2; 
               p[lastPtr][0] = prop->vertex[element][0]; 
               p[lastPtr][1] = prop->vertex[element][1];
               p[lastPtr][2] = prop->vertex[element][2];
               if (lastPtr == 1) {
                  stripReady = true;
               }
            }
         } else {
            // end of strip reached
            lastPtr = -1;
            stripReady = false;
         }
      }
   }

   rapidModel->EndModel();
}

//! destructor
Model::~Model() { }

