/* rom.c++  Range of motion simulation of artificial hip implants 
            Physical bases Simulation, ETH Zuerich SS03
            Rene Mueller, April 14th 2003 */

#include <iostream.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoSwitch.h>

#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoSphere.h>

#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/sensors/SoIdleSensor.h>
#include <Inventor/sensors/SoSensor.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>

#include "model.h"
#include "rom.h"
#include "collisionset.h"
#include "romanglepool.h"

#define DPHI 1.0
#define THETA_RESOLUTION .1

// Prototypes
void idle(void *data, SoSensor *s);
void keyCB(void *userData, SoEventCallback *eventCB);


//! root node of Inventor tree 
SoSeparator* root;

//! root node of Inventor movable nodes 
SoSeparator* movables;

//! collsion set that contains the model of all fixed elements (non-moving)
CollisionSet *fixed;

//! coolision set that contains the models of all movable elements
CollisionSet *movable;

//! Inventor translation for movable objects
SoTranslation* movableTranslation;

//! Inventor Rotation for movable objects
SoRotation* movableRotation;

//! data structore for holding the rom data 
RomAnglePool* romAnglePool;

//! true if romputation of rom data is not completed yet
bool firstPass = true;

//! booelan flag: true if ROM cone is visible 
bool visualisation = false;

//! switch node to enable/disable visualisation of ROM cone 
SoSwitch* visualisationSwitch;

//! boolean flag: true if axis are visiable
bool axisVisible = true;

//! switch node to enable/disable axis 
SoSwitch* axisSwitch;

//! phi  = orientation of rotation axis  
double phi = 0.0;

//! theta = elongation of leg
double theta = 0.0;

//! true if animation is running
bool running = false;


//! Main: this is where we start
int main(int , char **argv)
{
   Widget myWindow = SoXt::init(argv[0]);
   if (myWindow == NULL) exit(1);

   root = new SoSeparator;
   root->ref();
   
   // add key event listener
   SoIdleSensor *idleSensor = new SoIdleSensor;
   SoEventCallback *eventCB = new SoEventCallback;
   eventCB->addEventCallback(SoKeyboardEvent::getClassTypeId(), keyCB, idleSensor);
   root->addChild(eventCB);
   
   // select light source
   SoPointLight *light = new SoPointLight;
   light->location.setValue(100.0, 100.0, 100.0); 
   light->intensity = 0.5;
   light->on = true;
   root->addChild(light);
   
   Color diffuse, specular;

   axisSwitch = new SoSwitch;
   root->addChild(axisSwitch);
   axisSwitch->addChild(new SoSeparator); // child 0 = nothing (invisible)
   SoSeparator* axisNode = new SoSeparator;
   axisSwitch->addChild(axisNode);
   
   // draw xaxis
   SoCone *xcone = new SoCone;
   SoCylinder *xcyl = new SoCylinder;
   xcone->bottomRadius = 3.0;
   xcone->height = 12;
   xcyl->radius = 1.0;
   xcyl->height = 200;
   SoRotation *xrt = new SoRotation;
   SbVec3f axis;
   axis.setValue(0.0, 0.0, 1.0);
   xrt->rotation.setValue(axis, -M_PI/2.0);
   SoTranslation *xtr1 = new SoTranslation;
   xtr1->translation.setValue(0,100,0);
   SoTranslation *xtr2 = new SoTranslation;
   xtr2->translation.setValue(0, 106, 0);
   SoMaterial *xmat = new SoMaterial;
   xmat->diffuseColor.setValue(1.0, 0.0, 0.0);
   xmat->specularColor.setValue(1.0, 1.0, 1.0);
   xmat->ambientColor.setValue(0.0, 0.0, 0.0);
   SoSeparator *xaxis = new SoSeparator;
   axisNode->addChild(xaxis);
   xaxis->addChild(xmat);
   xaxis->addChild(xrt);
   xaxis->addChild(xtr1);
   xaxis->addChild(xcyl);
   xaxis->addChild(xtr2);
   xaxis->addChild(xcone);
   
   // draw yaxis
   SoCone *ycone = new SoCone;
   SoCylinder *ycyl = new SoCylinder;
   ycone->bottomRadius = 3.0;
   ycone->height = 12;
   ycyl->radius = 1.0;
   ycyl->height = 200;
   SoTranslation *ytr1 = new SoTranslation;
   ytr1->translation.setValue(0,100,0);
   SoTranslation *ytr2 = new SoTranslation;
   ytr2->translation.setValue(0, 106, 0);
   SoMaterial *ymat = new SoMaterial;
   ymat->diffuseColor.setValue(0.0, 1.0, 0.0);
   ymat->specularColor.setValue(1.0, 1.0, 1.0);
   ymat->ambientColor.setValue(0.0, 0.0, 0.0);
   SoSeparator *yaxis = new SoSeparator;
   axisNode->addChild(yaxis);
   yaxis->addChild(ymat);
   yaxis->addChild(ytr1);
   yaxis->addChild(ycyl);
   yaxis->addChild(ytr2);
   yaxis->addChild(ycone);
  
   // draw zaxis
   SoCone *zcone = new SoCone;
   SoCylinder *zcyl = new SoCylinder;
   zcone->bottomRadius = 3.0;
   zcone->height = 12;
   zcyl->radius = 1.0;
   zcyl->height = 200;
   SoRotation *zrt = new SoRotation;
   axis.setValue(1.0, 0.0, 0.0);
   zrt->rotation.setValue(axis, M_PI/2.0);
   SoTranslation *ztr1 = new SoTranslation;
   ztr1->translation.setValue(0, 100, 0);
   SoTranslation *ztr2 = new SoTranslation;
   ztr2->translation.setValue(0, 106, 0);
   SoMaterial *zmat = new SoMaterial;
   zmat->diffuseColor.setValue(0.0, 0.0, 1.0);
   zmat->specularColor.setValue(1.0, 1.0, 1.0);
   zmat->ambientColor.setValue(0.0, 0.0, 0.0);
   SoSeparator *zaxis = new SoSeparator;
   axisNode->addChild(zaxis);
   zaxis->addChild(zmat);
   zaxis->addChild(zrt);
   zaxis->addChild(ztr1);
   zaxis->addChild(zcyl);
   zaxis->addChild(ztr2);
   zaxis->addChild(zcone);

   // axis is initially visible 
   axisSwitch->whichChild = 1;   
   
   /* open fixed objects */
   // read antilux 
   Model *antilux = new Model(ANTILUX_MODEL);
   antilux->getMaterial()->diffuseColor.setValue(0.1, 0.5, 0.0);
   antilux->getMaterial()->specularColor.setValue(0.2, 0.8, 0.0);
   antilux->getMaterial()->shininess = 0.5;
   root->addChild(antilux->getRootModel());
   
   // read bicon 
   Model *bicon = new Model(BICON_MODEL);
   bicon->getMaterial()->diffuseColor.setValue(0.5, 0.1, 0.0);
   bicon->getMaterial()->specularColor.setValue(0.8, 0.2, 0.0);
   bicon->getMaterial()->shininess = 0.5;
   root->addChild(bicon->getRootModel());

   // read hip
   Model *hip = new Model(HIP_MODEL);
   hip->getMaterial()->diffuseColor.setValue(0.6, 0.5, 0.3);
   hip->getMaterial()->specularColor.setValue(0.0, 0.0, 0.0);
   hip->getMaterial()->shininess = 0.5;
   root->addChild(hip->getRootModel());

   /* open movable objects */
   movables = new SoSeparator;
   movableTranslation = new SoTranslation;
   movableRotation = new SoRotation;
   movables->addChild(movableTranslation);
   movables->addChild(movableRotation);
   root->addChild(movables);

   // read shaft
   Model *shaft = new Model(SHAFT_MODEL);
   shaft->getMaterial()->diffuseColor.setValue(0.5, 0.5, 0.5);
   shaft->getMaterial()->specularColor.setValue(0.7, 0.7, 0.7);
   shaft->getMaterial()->shininess = 0.5;
   movables->addChild(shaft->getRootModel());

   // read femur
   Model *femur = new Model(FEMUR_MODEL);
   femur->getMaterial()->diffuseColor.setValue(0.6, 0.5, 0.3);
   femur->getMaterial()->specularColor.setValue(0.0, 0.0, 0.0);
   femur->getMaterial()->shininess = 0.99;
   movables->addChild(femur->getRootModel());
   
   // set up collision sets
   fixed = new CollisionSet;
   fixed->insert(hip);
   fixed->insert(antilux);
   fixed->insert(bicon);
   
   movable = new CollisionSet;
   movable->insert(femur);
   movable->insert(shaft);

   romAnglePool = new RomAnglePool;

   
   //! setup cone material
   SoMaterial *conemat = new SoMaterial;
   conemat->diffuseColor.setValue(1.0, 0.0, 0.0);
   conemat->specularColor.setValue(1.0, 1.0, 1.0);
   conemat->ambientColor.setValue(0.0, 0.0, 0.0);
   root->addChild(conemat);
   visualisationSwitch = new SoSwitch;
   root->addChild(visualisationSwitch);
   visualisationSwitch->addChild(new SoSeparator); // child 0 is dummy 
   visualisationSwitch->whichChild = 0;
   

   // setup idle sensor
   idleSensor->setFunction(idle);
   idleSensor->setData(root);
   if (running) {
      idleSensor->schedule();
   }

   // Set up viewer:
   SoXtExaminerViewer *myViewer = 
            new SoXtExaminerViewer(myWindow);
   myViewer->setSceneGraph(root);
   myViewer->setTitle("Examiner Viewer");
   myViewer->setBackgroundColor(SbColor(0.0, 0.0, 0.3));
   //myViewer->setViewing(false);

   myViewer->show();

   SoXt::show(myWindow);
   SoXt::mainLoop();
}


//! callback function for Idle-Sensor
/*!
 *  This callback function is called whenever Openinventor is in
 *  idle state. During idle state the next simulation step is
 *  computed.
 *  \param data user data (in this case reference to model)
 *  \param sensor reference to sensor instance that triggered event
 */
void idle(void *data, SoSensor *sensor)
{
   double step = 90;
   theta = 0; 
   SbVec3f axis;

   while (step > THETA_RESOLUTION/2.0) { 
      step = step/2.0;
  
      // compute new elongation
      double ux = 0;   // normalized rotation axis (u)
      double uy = -sin(phi*M_PI/180.0); 
      double uz = -cos(phi*M_PI/180.0);
      axis.setValue(ux, uy, uz);
      
      // trace positioning
      // movableRotation->rotation.setValue(axis, theta*M_PI/180.0);
   
      // compute rotation matrix of rotation using quaterions
      // Q = ( q0, (q1, q2, q3) )
      // u = axis vector (normalized)
      // theta = rotation angle
      double q0 = cos(theta*M_PI/360.0);      // q0 = cos(theta/2.0)
      double s = sin(theta*M_PI/360.0);       
      double q1 = ux*s;    // q1 = ux * sin(theta/2)
      double q2 = uy*s;    // q2 = uy * sin(theta/2)
      double q3 = uz*s;    // q3 = uz * sin(theta/2)

      // translate quaterion back into matrix
      double R[3][3];
      R[0][0] = 1 - 2*q2*q2 - 2*q3*q3;
      R[0][1] = 2*q1*q2 - 2*q0*q3;
      R[0][2] = 2*q1*q3 + 2*q0*q2 ; 
      R[1][0] = 2*q1*q2 + 2*q0*q3; 
      R[1][1] = 1 - 2*q1*q1 - 2*q3*q3;
      R[1][2] = 2*q2*q3 - 2*q0*q1; 
      R[2][0] = 2*q1*q3 - 2*q0*q2;
      R[2][1] = 2*q2*q3 + 2*q0*q1;
      R[2][2] = 1 - 2*q1*q1 - 2*q2*q2;
      movable->setOrientation(R);
   
      if (CollisionSet::checkForCollision(fixed, movable)) {
         theta += -step;   // collision
      } else {
         theta += step;    // no collision
      }
   }
   
   // only display end position
   movableRotation->rotation.setValue(axis, theta*M_PI/180.0);
   if (firstPass) {
      // add rom data into rom angle pool
      romAnglePool->append(phi,theta);
   }
         
   // next phi
   phi = phi+DPHI;
   if (phi >= 360) {
      phi = phi-360;

      if (firstPass) {
         firstPass = false;
         visualisationSwitch->addChild(romAnglePool->createVisualizationCone());
         if (visualisation) {
            visualisationSwitch->whichChild = 1;
         }
      }
   }
   if (running) {
      sensor->schedule();   
   }
}

//! callback function for keyboard event
/*!
 *  This callback is called whenever the user hits a key.
 *  It toggles the cone visualisation when 'C' is pressed.
 *  \param userData user (NULL in this case)
 *  \param eventCB event callback that was triggered
 */
void keyCB(void *userData, SoEventCallback *eventCB)
{
   SoIdleSensor* idleSensor = (SoIdleSensor*)(userData);
   if (SO_KEY_PRESS_EVENT(eventCB->getEvent(), SoKeyboardEvent::C)) {
      visualisation = !visualisation;
      if (visualisation) {
         visualisationSwitch->whichChild = 1;
      } else {
         visualisationSwitch->whichChild = 0;
      }
   } else if (SO_KEY_PRESS_EVENT(eventCB->getEvent(), SoKeyboardEvent::A)) {
      axisVisible = !axisVisible;
      if (axisVisible) {
         axisSwitch->whichChild = 1; 
      } else {
         axisSwitch->whichChild = 0;
      }
   } else if (SO_KEY_PRESS_EVENT(eventCB->getEvent(), SoKeyboardEvent::S)) {
      running = !running;
      if (running) {
         idleSensor->schedule();
      } 
   } 
}
